#define STRUCT privcmd_hypercall_t
static void
print_hypercall_enter(print_status *status, const privcmd_hypercall_t *data)
{
	const base_type* type = NULL;
	const char *name = NULL;
	int have_args = 1;

#define HYPERCALL_NOARG(_name) case XENMEM_ ## _name: \
	name = #_name; \
	have_args = 0; \
	break;
#define HYPERCALL(_name, struct, item) case XENMEM_ ## _name: \
	name = #_name; \
	type = &type_ ## struct.base; \
	break;
#define HYPERCALL_OUT(name, struct, item) HYPERCALL_NOARG(name)

	switch (data->arg[0]) {
	HYPERCALL(increase_reservation, xen_memory_reservation, reservation) // extent_start out
	HYPERCALL(decrease_reservation, xen_memory_reservation, reservation) // extent_start in
	HYPERCALL_NOARG(maximum_ram_page)
	HYPERCALL(current_reservation, domid, domid)
	HYPERCALL(maximum_reservation, domid, domid)
	HYPERCALL_OUT(machphys_mfn_list, xen_machphys_mfn_list, machphys_mfn_list)	// mostly out
	HYPERCALL(populate_physmap, xen_memory_reservation, reservation)	// extent_start in/out
	HYPERCALL(add_to_physmap, xen_add_to_physmap, add_to_physmap)	// input ??
	HYPERCALL_OUT(memory_map, xen_memory_map, memory_map)	// some input!
	HYPERCALL_OUT(machine_memory_map, xen_memory_map, memory_map)	// like memory_map
	HYPERCALL(exchange, xen_memory_exchange, exchange)	// some in some out
	HYPERCALL_OUT(machphys_mapping, xen_machphys_mapping, machphys_mapping)
	HYPERCALL(set_memory_map, xen_foreign_memory_map, foreign_memory_map)
	HYPERCALL_NOARG(maximum_gpfn)
	// defined in memory but no union for it in privcmd
//	HYPERCALL(set_pod_target, xen_pod_target, )	// input
//	HYPERCALL_OUT(get_pod_target, xen_pod_target, )
	HYPERCALL_NOARG(get_sharing_freed_pages)
#undef HYPERCALL_NOARG
#undef HYPERCALL
#undef HYPERCALL_OUT
	}

	status_printf(status, "{op=memory_op,");
	if (name)
		status_printf(status, "cmd=%s", name);
	else
		status_printf(status, "cmd=UNKNOWN(%lu)", data->arg[0]);
	status_printf(status, have_args ? "," : "}");

	if (type) {
		char buf[1024];
		if (type->size > sizeof(buf) || umoven(status->tcp, data->arg[1], type->size, buf) < 0)
	                status_printf(status, "BADPTR:%#lx", (long) data->arg[1]);
		else
			type->print(status, buf, type);
		if (have_args)
			status_printf(status, "}");
	}
}

static void
print_hypercall_exit(print_status *status, const privcmd_hypercall_t *data)
{
	const base_type* type = NULL;

#define HYPERCALL_NOARG(_name) case XENMEM_ ## _name: \
	break;
// print just end
#define HYPERCALL(name, struct, item) HYPERCALL_NOARG(name)
// print
#define HYPERCALL_OUT(_name, struct, item) case XENMEM_ ## _name: \
	type = &type_ ## struct.base; \
	break;

	switch (data->arg[0]) {
	HYPERCALL(increase_reservation, xen_memory_reservation, reservation) // extent_start out
	HYPERCALL(decrease_reservation, xen_memory_reservation, reservation) // extent_start in
	HYPERCALL_NOARG(maximum_ram_page)
	HYPERCALL(current_reservation, domid, domid)
	HYPERCALL(maximum_reservation, domid, domid)
	HYPERCALL_OUT(machphys_mfn_list, xen_machphys_mfn_list, machphys_mfn_list)	// mostly out
	HYPERCALL(populate_physmap, xen_memory_reservation, reservation)	// extent_start in/out
	HYPERCALL(add_to_physmap, xen_add_to_physmap, add_to_physmap)	// input ??
	HYPERCALL_OUT(memory_map, xen_memory_map, memory_map)	// some input!
	HYPERCALL_OUT(machine_memory_map, xen_memory_map, memory_map)	// like memory_map
	HYPERCALL(exchange, xen_memory_exchange, exchange)	// some in some out
	HYPERCALL_OUT(machphys_mapping, xen_machphys_mapping, machphys_mapping)
	HYPERCALL(set_memory_map, xen_foreign_memory_map, foreign_memory_map)
	HYPERCALL_NOARG(maximum_gpfn)
	// defined in memory but no union for it in privcmd
//	HYPERCALL(set_pod_target, xen_pod_target, )	// input
//	HYPERCALL_OUT(get_pod_target, xen_pod_target, )
	HYPERCALL_NOARG(get_sharing_freed_pages)
#undef HYPERCALL_NOARG
#undef HYPERCALL
#undef HYPERCALL_OUT
	}

	if (type) {
		char buf[1024];
		if (type->size > sizeof(buf) || umoven(status->tcp, data->arg[1], type->size, buf) < 0)
	                status_printf(status, "BADPTR:%#lx", (long) data->arg[1]);
		else
			type->print(status, buf, type);
	}

	status_printf(status, "}");
}

static void
print_hypercall(print_status *status, const void *_data, const void *dummy)
{
	const privcmd_hypercall_t *data = (const privcmd_hypercall_t *)_data;

	if (data->op == __HYPERVISOR_memory_op) {
		if (status->entering)
			print_hypercall_enter(status, data);
		else
			print_hypercall_exit(status, data);
	} else {
		type_privcmd_hypercall.base.print(status, _data, &type_privcmd_hypercall.base);
	}
}

static const type_base type_hypercall = {
	{ print_hypercall, sizeof(STRUCT) }
};
#undef STRUCT


