/*
 * Copyright (c) 2014 Citrix.
 * Written by Frediano Ziglio <frediano.ziglio@citrix.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if EXACT_TESTING
#  include "mock_defs.h"
#else
#  include "defs.h"
#endif

#include <stddef.h>
#include <stdarg.h>

#include <sys/ioctl.h>
#include <xenctrl.h>
#include <xen/sys/evtchn.h>
#include <xen/sys/privcmd.h>
#include <xen/domctl.h>
#include <xen/sched.h>

#include "structs.h"
#include "structures.h"

// TODO finish
#ifdef IOCTL_PRIVCMD_HVMOP

#define STRUCT privcmd_hvmop_t
static const switch_item privcmd_hvmop_cmd_items[] = {
#define HVMOP(name, type, output) \
	{ HVMOP_ ## name, #name, &type_ ## type.base, output },
	HVMOP(set_param, xen_hvm_param, 0)
// TODO output !!!
	HVMOP(get_param, xen_hvm_param, 1)
	/* HVMOP(set_pci_intx_level, xen_hvm_set_pci_intx_level, 0) */
	/* HVMOP(set_isa_irq_level, xen_hvm_set_isa_irq_level, 0) */
	/* HVMOP(set_pci_link_route, xen_hvm_set_pci_link_route, 0) */
	HVMOP(flush_tlbs, void, 0)
// HVMOP_track_dirty_vram
	/* HVMOP(modified_memory, xen_hvm_modified_memory, 0) */
	/* HVMOP(set_mem_type, xen_hvm_set_mem_type, 0) */
#ifdef HVMOP_process_dying
	HVMOP(process_dying, void, 0)
#endif
	HVMOP(pagetable_dying, xen_hvm_pagetable_dying, 0)
// TODO output !!!
	HVMOP(get_time, xen_hvm_get_time, 1)
	HVMOP(xentrace, xen_hvm_xentrace, 0)
	HVMOP(set_mem_access, xen_hvm_set_mem_access, 0)
// TODO output !!!
	HVMOP(get_mem_access, xen_hvm_get_mem_access, 1)
	/* HVMOP(inject_trap, xen_hvm_inject_trap, 0) */
// TODO output !!!
	HVMOP(get_mem_type, xen_hvm_get_mem_type, 1)
#ifdef HVMOP_set_driver_version
	HVMOP(set_driver_version, xen_hvm_set_driver_version, 0)
#endif
	{ 0, NULL, }
};

static const type_switch type_privcmd_hvmop_cmd = {
	{ print_switch, sizeof(unsigned) },
	privcmd_hvmop_cmd_items,
	offsetof(privcmd_hvmop_t, u) - offsetof(privcmd_hvmop_t, cmd)
};
START_STRUCT(privcmd_hvmop)
	BASE_ITEM(privcmd_hvmop_cmd, cmd)
END_STRUCT(privcmd_hvmop)
#undef STRUCT
#endif

#if 0
#define STRUCT privcmd_memop_t
static switch_item privcmd_memop_cmd_items[] = {
	{ XENMEM_increase_reservation,   "increase_reservation",   &type_reservation },
	{ XENMEM_decrease_reservation,   "decrease_reservation",   &type_reservation }, 
	{ XENMEM_populate_physmap,       "populate_physmap",       &type_reservation },
	{ XENMEM_exchange,               "exchange",               &type_exchange },
	{ XENMEM_maximum_gpfn,           "maximum_gpfn",           &type_void },
	{ XENMEM_maximum_ram_page,       "maximum_ram_page",       &type_domid },
	{ XENMEM_current_reservation,    "current_reservation",    &type_domid },
	{ XENMEM_maximum_reservation,    "maximum_reservation",    &type_domid },
	{ XENMEM_add_to_physmap,         "add_to_physmap",         &type_add_to_physmap },
	{ XENMEM_set_memory_map,         "set_memory_map",         &type_foreign_memory_map },
	{ XENMEM_machphys_mfn_list,      "machphys_mfn_list",      &type_machphys_mfn_list },
	{ XENMEM_machphys_mapping,       "machphys_mapping",       &type_machphys_mapping },
	{ XENMEM_memory_map,             "memory_map",             &type_memory_map },
	{ XENMEM_machine_memory_map,     "machine_memory_map",     &type_memory_map },
	{ XENMEM_get_sharing_freed_pages,"get_sharing_freed_pages",&type_void },
// XENMEM_set_pod_target and XENMEM_get_pod_target ??
	{ 0, NULL, }
};
static const type_switch type_privcmd_memop_cmd = {
	print_switch,
	sizeof(unsigned),
	privcmd_memop_cmd_items,
	offsetof(privcmd_memop_t, u) - offsetof(privcmd_memop_t, cmd)
};
START_STRUCT(privcmd_memop)
	BASE_ITEM(privcmd_memop_cmd, cmd)
END_STRUCT(privcmd_memop)
#undef STRUCT
#endif

static void
print_xen_domctl_bind_pt_irq(print_status *status, const void *_data, const void *dummy)
{
	const struct xen_domctl_bind_pt_irq *data = (const struct xen_domctl_bind_pt_irq*)_data;

	status_printf(status, "{machine_irq=%u,", data->machine_irq);
	status_printf(status, "irq_type=%u,", data->irq_type);
#if XEN_DOMCTL_INTERFACE_VERSION < 0x0000000e
	status_printf(status, "hvm_domid=%u,", data->hvm_domid);
#endif
	switch (data->irq_type) {
	case PT_IRQ_TYPE_PCI:
		status_printf(status, "{pci=%u,%u,%u}", data->u.pci.bus, data->u.pci.device, data->u.pci.intx);
		break;
	case PT_IRQ_TYPE_ISA:
		status_printf(status, "{isa=%u}", data->u.isa.isa_irq);
		break;
	case PT_IRQ_TYPE_MSI:
	case PT_IRQ_TYPE_MSI_TRANSLATE:
		status_printf(status, "{msi=%u,%u", data->u.msi.gvec, data->u.msi.gflags);
		status_printf(status, "%" PRIu64 "}", data->u.msi.gtable);
		break;
	}
	status_printf(status, "}");
}

static const type_base type_xen_domctl_bind_pt_irq = {
	{ print_xen_domctl_bind_pt_irq, sizeof(struct xen_domctl_bind_pt_irq) }
};

#define STRUCT struct xen_domctl
static void
print_xen_domctl_enter(print_status *status, const struct xen_domctl *data)
{
	const base_type* type = NULL;
	const char *name = NULL;

#define XEN_DOMCTL_NOARG(_name) case XEN_DOMCTL_ ## _name: \
	name = #_name; \
	break;
#define XEN_DOMCTL(_name, struct, item) case XEN_DOMCTL_ ## _name: \
	name = #_name; \
	type = &type_ ## struct.base; \
	break;
#define XEN_DOMCTL_OUT(name, struct, item) XEN_DOMCTL_NOARG(name)

	switch (data->cmd) {
	XEN_DOMCTL(createdomain, xen_domctl_createdomain, createdomain)	// .domain output
	XEN_DOMCTL_NOARG(destroydomain)
	XEN_DOMCTL_NOARG(pausedomain)
	XEN_DOMCTL_NOARG(unpausedomain)
	XEN_DOMCTL_OUT(getdomaininfo, xen_domctl_getdomaininfo, getdomaininfo)	// output
#ifdef XEN_DOMCTL_getpageframeinfo
	XEN_DOMCTL_OUT(getpageframeinfo, xen_domctl_getpageframeinfo, getpageframeinfo)	// output, gmfn in
#endif
#ifdef XEN_DOMCTL_getpageframeinfo2
	XEN_DOMCTL_OUT(getpageframeinfo2, xen_domctl_getpageframeinfo2, getpageframeinfo2)	// num, array input, array output
#endif
	XEN_DOMCTL(setvcpuaffinity, xen_domctl_vcpuaffinity, vcpuaffinity)	// input
//	XEN_DOMCTL(shadow_op, xen_domctl_shadow_op, shadow_op)	// input/output ??
	XEN_DOMCTL(max_mem, xen_domctl_max_mem, max_mem)	// input, (set mem)
	XEN_DOMCTL(setvcpucontext, xen_domctl_vcpucontext, vcpucontext)	// input
	XEN_DOMCTL_OUT(getvcpucontext, xen_domctl_vcpucontext, vcpucontext)	// vcpu in output
	XEN_DOMCTL_OUT(getvcpuinfo, xen_domctl_getvcpuinfo, getvcpuinfo)	// vcpu in, output
#ifdef XEN_DOMCTL_max_vcpus
	XEN_DOMCTL(max_vcpus, xen_domctl_max_vcpus, max_vcpus)		// input
#endif
//	XEN_DOMCTL(scheduler_op, xen_domctl_scheduler_op, scheduler_op)	// depends ??
	XEN_DOMCTL(setdomainhandle, xen_domctl_setdomainhandle, setdomainhandle)	// input
	XEN_DOMCTL(setdebugging, xen_domctl_setdebugging, setdebugging)	// input
	XEN_DOMCTL(irq_permission, xen_domctl_irq_permission, irq_permission)	// input
	XEN_DOMCTL(iomem_permission, xen_domctl_iomem_permission, iomem_permission)	// input
	XEN_DOMCTL(ioport_permission, xen_domctl_ioport_permission, ioport_permission)	// input
	XEN_DOMCTL(hypercall_init, xen_domctl_hypercall_init, hypercall_init)	// input
//	XEN_DOMCTL(arch_setup, xen_domctl_arch_setup, arch_setup)		// only ia64
	XEN_DOMCTL(settimeoffset, xen_domctl_settimeoffset, settimeoffset)	// input
	XEN_DOMCTL_OUT(getvcpuaffinity, xen_domctl_vcpuaffinity, vcpuaffinity)	// output
//	XEN_DOMCTL(real_mode_area, xen_domctl_real_mode_area, real_mode_area)	// unused!
	XEN_DOMCTL_NOARG(resumedomain)
	XEN_DOMCTL(sendtrigger, xen_domctl_sendtrigger, sendtrigger)	// input
	XEN_DOMCTL(subscribe, xen_domctl_subscribe, subscribe)	// input
	XEN_DOMCTL_OUT(gethvmcontext, xen_domctl_hvmcontext, hvmcontext)	// output
	XEN_DOMCTL(sethvmcontext, xen_domctl_hvmcontext, hvmcontext)	// input
	XEN_DOMCTL(set_address_size, xen_domctl_address_size, address_size)	// input
	XEN_DOMCTL_OUT(get_address_size, xen_domctl_address_size, address_size)	// output
	XEN_DOMCTL(assign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(bind_pt_irq, xen_domctl_bind_pt_irq, bind_pt_irq)	// input ??
	XEN_DOMCTL(memory_mapping, xen_domctl_memory_mapping, memory_mapping)	// input
	XEN_DOMCTL(ioport_mapping, xen_domctl_ioport_mapping, ioport_mapping)	// input
#ifdef XEN_DOMCTL_pin_mem_cacheattr
	XEN_DOMCTL(pin_mem_cacheattr, xen_domctl_pin_mem_cacheattr, pin_mem_cacheattr)	// input
#endif
	XEN_DOMCTL(set_ext_vcpucontext, xen_domctl_ext_vcpucontext, ext_vcpucontext)	// input
	XEN_DOMCTL_OUT(get_ext_vcpucontext, xen_domctl_ext_vcpucontext, ext_vcpucontext)	// output, vcpu input
//	XEN_DOMCTL(set_opt_feature, xen_domctl_set_opt_feature, set_opt_feature)	// only ia64
	XEN_DOMCTL(test_assign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(set_target, xen_domctl_set_target, set_target)	// input
	XEN_DOMCTL(deassign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(unbind_pt_irq, xen_domctl_bind_pt_irq, bind_pt_irq)	// input
#if defined(__i386__) || defined(__x86_64__)
#ifdef XEN_DOMCTL_set_cpuid
	XEN_DOMCTL(set_cpuid, xen_domctl_cpuid, cpuid)	// input
#endif
#endif
	XEN_DOMCTL_OUT(get_device_group, xen_domctl_get_device_group, get_device_group)	// output, machine_bdf, max_sdevs input
#ifdef XEN_DOMCTL_set_machine_address_size
	XEN_DOMCTL(set_machine_address_size, xen_domctl_address_size, address_size)	// input
#endif
#ifdef XEN_DOMCTL_get_machine_address_size
	XEN_DOMCTL_OUT(get_machine_address_size, xen_domctl_address_size, address_size)	// output
#endif
#ifdef XEN_DOMCTL_suppress_spurious_page_faults
	XEN_DOMCTL_NOARG(suppress_spurious_page_faults)
#endif
	XEN_DOMCTL(debug_op, xen_domctl_debug_op, debug_op)	// input
	XEN_DOMCTL_OUT(gethvmcontext_partial, xen_domctl_hvmcontext_partial, hvmcontext_partial)	// output, type, instance input
#ifdef XEN_DOMCTL_MEM_EVENT_OP_SHARING
	XEN_DOMCTL(mem_event_op, xen_domctl_mem_event_op, mem_event_op)	// input
#endif
#ifdef XEN_DOMCTL_VM_EVENT_OP_SHARING
	XEN_DOMCTL(vm_event_op, xen_domctl_vm_event_op, vm_event_op)	// input
#endif
//	XEN_DOMCTL(mem_sharing_op, xen_domctl_mem_sharing_op, mem_sharing_op)	// depends
#if defined(XEN_DOMCTL_disable_migrate)
	XEN_DOMCTL(disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// input
#elif defined(XEN_DOMCTL_set_disable_migrate)
	XEN_DOMCTL(set_disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// input
#endif
#ifdef XEN_DOMCTL_query_disable_migrate
	XEN_DOMCTL_OUT(query_disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// output
#endif
	XEN_DOMCTL_OUT(gettscinfo, xen_domctl_tsc_info, tsc_info)	// output
	XEN_DOMCTL(settscinfo, xen_domctl_tsc_info, tsc_info)	// input
	// TODO handle input and output of same field
	XEN_DOMCTL(getpageframeinfo3, xen_domctl_getpageframeinfo3, getpageframeinfo3)	// num, array input, array output
	XEN_DOMCTL(setvcpuextstate, xen_domctl_vcpuextstate, vcpuextstate)	// input
	XEN_DOMCTL_OUT(getvcpuextstate, xen_domctl_vcpuextstate, vcpuextstate)	// output, vcpu, size, xfeature_mask input
	XEN_DOMCTL(set_access_required, xen_domctl_set_access_required, access_required)	// input
	XEN_DOMCTL_OUT(get_runstate_info, xen_domctl_runstate_info, domain_runstate)	// output
	XEN_DOMCTL(setcorespersocket, xen_domctl_corespersocket, corespersocket)	// input
	XEN_DOMCTL_OUT(gdbsx_guestmemio, xen_domctl_gdbsx_memio, gdbsx_guest_memio)	// see structure, mostly input
	XEN_DOMCTL(gdbsx_pausevcpu, xen_domctl_gdbsx_pauseunp_vcpu, gdbsx_pauseunp_vcpu)	// input
	XEN_DOMCTL(gdbsx_unpausevcpu, xen_domctl_gdbsx_pauseunp_vcpu, gdbsx_pauseunp_vcpu)	// input
	XEN_DOMCTL_OUT(gdbsx_domstatus, xen_domctl_gdbsx_domstatus, gdbsx_domstatus)	// output
#undef XEN_DOMCTL_NOARG
#undef XEN_DOMCTL
#undef XEN_DOMCTL_OUT
	}

	if (name)
		status_printf(status, "{cmd=%s", name);
	else
		status_printf(status, "{cmd=UNKNOWN(%lu)", data->cmd);

	status_printf(status, ",dom=%u,int=%u,", data->domain, (unsigned) data->interface_version);

	if (type)
		type->print(status, &data->u, type);
}

static void
print_xen_domctl_exit(print_status *status, const struct xen_domctl *data)
{
	const base_type* type = NULL;

#define XEN_DOMCTL_NOARG(_name) case XEN_DOMCTL_ ## _name: \
	break;
// print just end
#define XEN_DOMCTL(name, struct, item) XEN_DOMCTL_NOARG(name)
// print
#define XEN_DOMCTL_OUT(_name, struct, item) case XEN_DOMCTL_ ## _name: \
	type = &type_ ## struct.base; \
	break;

	switch (data->cmd) {
	XEN_DOMCTL(createdomain, xen_domctl_createdomain, createdomain)	// .domain output
	XEN_DOMCTL_NOARG(destroydomain)
	XEN_DOMCTL_NOARG(pausedomain)
	XEN_DOMCTL_NOARG(unpausedomain)
	XEN_DOMCTL_OUT(getdomaininfo, xen_domctl_getdomaininfo, getdomaininfo)	// output
#ifdef XEN_DOMCTL_getpageframeinfo
	XEN_DOMCTL_OUT(getpageframeinfo, xen_domctl_getpageframeinfo, getpageframeinfo)	// output, gmfn in
#endif
#ifdef XEN_DOMCTL_getpageframeinfo2
	XEN_DOMCTL_OUT(getpageframeinfo2, xen_domctl_getpageframeinfo2, getpageframeinfo2)	// num, array input, array output
#endif
	XEN_DOMCTL(setvcpuaffinity, xen_domctl_vcpuaffinity, vcpuaffinity)	// input
//	XEN_DOMCTL(shadow_op, xen_domctl_shadow_op, shadow_op)	// input/output ??
	XEN_DOMCTL(max_mem, xen_domctl_max_mem, max_mem)	// input, (set mem)
	XEN_DOMCTL(setvcpucontext, xen_domctl_vcpucontext, vcpucontext)	// input
	XEN_DOMCTL_OUT(getvcpucontext, xen_domctl_vcpucontext, vcpucontext)	// vcpu in output
	XEN_DOMCTL_OUT(getvcpuinfo, xen_domctl_getvcpuinfo, getvcpuinfo)	// vcpu in, output
	XEN_DOMCTL(max_vcpus, xen_domctl_max_vcpus, max_vcpus)		// input
//	XEN_DOMCTL(scheduler_op, xen_domctl_scheduler_op, scheduler_op)	// depends ??
	XEN_DOMCTL(setdomainhandle, xen_domctl_setdomainhandle, setdomainhandle)	// input
	XEN_DOMCTL(setdebugging, xen_domctl_setdebugging, setdebugging)	// input
	XEN_DOMCTL(irq_permission, xen_domctl_irq_permission, irq_permission)	// input
	XEN_DOMCTL(iomem_permission, xen_domctl_iomem_permission, iomem_permission)	// input
	XEN_DOMCTL(ioport_permission, xen_domctl_ioport_permission, ioport_permission)	// input
	XEN_DOMCTL(hypercall_init, xen_domctl_hypercall_init, hypercall_init)	// input
//	XEN_DOMCTL(arch_setup, xen_domctl_arch_setup, arch_setup)		// only ia64
	XEN_DOMCTL(settimeoffset, xen_domctl_settimeoffset, settimeoffset)	// input
	XEN_DOMCTL_OUT(getvcpuaffinity, xen_domctl_vcpuaffinity, vcpuaffinity)	// output
//	XEN_DOMCTL(real_mode_area, xen_domctl_real_mode_area, real_mode_area)	// unused!
	XEN_DOMCTL_NOARG(resumedomain)
	XEN_DOMCTL(sendtrigger, xen_domctl_sendtrigger, sendtrigger)	// input
	XEN_DOMCTL(subscribe, xen_domctl_subscribe, subscribe)	// input
	XEN_DOMCTL_OUT(gethvmcontext, xen_domctl_hvmcontext, hvmcontext)	// output
	XEN_DOMCTL(sethvmcontext, xen_domctl_hvmcontext, hvmcontext)	// input
	XEN_DOMCTL(set_address_size, xen_domctl_address_size, address_size)	// input
	XEN_DOMCTL_OUT(get_address_size, xen_domctl_address_size, address_size)	// output
	XEN_DOMCTL(assign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(bind_pt_irq, xen_domctl_bind_pt_irq, bind_pt_irq)	// input ??
	XEN_DOMCTL(memory_mapping, xen_domctl_memory_mapping, memory_mapping)	// input
	XEN_DOMCTL(ioport_mapping, xen_domctl_ioport_mapping, ioport_mapping)	// input
#ifdef XEN_DOMCTL_pin_mem_cacheattr
	XEN_DOMCTL(pin_mem_cacheattr, xen_domctl_pin_mem_cacheattr, pin_mem_cacheattr)	// input
#endif
	XEN_DOMCTL(set_ext_vcpucontext, xen_domctl_ext_vcpucontext, ext_vcpucontext)	// input
	XEN_DOMCTL_OUT(get_ext_vcpucontext, xen_domctl_ext_vcpucontext, ext_vcpucontext)	// output, vcpu input
//	XEN_DOMCTL(set_opt_feature, xen_domctl_set_opt_feature, set_opt_feature)	// only ia64
	XEN_DOMCTL(test_assign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(set_target, xen_domctl_set_target, set_target)	// input
	XEN_DOMCTL(deassign_device, xen_domctl_assign_device, assign_device)	// input
	XEN_DOMCTL(unbind_pt_irq, xen_domctl_bind_pt_irq, bind_pt_irq)	// input
#if defined(__i386__) || defined(__x86_64__)
#ifdef XEN_DOMCTL_set_cpuid
	XEN_DOMCTL(set_cpuid, xen_domctl_cpuid, cpuid)	// input
#endif
#endif
	XEN_DOMCTL_OUT(get_device_group, xen_domctl_get_device_group, get_device_group)	// output, machine_bdf, max_sdevs input
#ifdef XEN_DOMCTL_set_machine_address_size
	XEN_DOMCTL(set_machine_address_size, xen_domctl_address_size, address_size)	// input
#endif
#ifdef XEN_DOMCTL_get_machine_address_size
	XEN_DOMCTL_OUT(get_machine_address_size, xen_domctl_address_size, address_size)	// output
#endif
#ifdef XEN_DOMCTL_suppress_spurious_page_faults
	XEN_DOMCTL_NOARG(suppress_spurious_page_faults)
#endif
	XEN_DOMCTL(debug_op, xen_domctl_debug_op, debug_op)	// input
	XEN_DOMCTL_OUT(gethvmcontext_partial, xen_domctl_hvmcontext_partial, hvmcontext_partial)	// output, type, instance input
#ifdef XEN_DOMCTL_MEM_EVENT_OP_SHARING
	XEN_DOMCTL(mem_event_op, xen_domctl_mem_event_op, mem_event_op)	// input
#endif
#ifdef XEN_DOMCTL_VM_EVENT_OP_SHARING
	XEN_DOMCTL(vm_event_op, xen_domctl_vm_event_op, vm_event_op)	// input
#endif
//	XEN_DOMCTL(mem_sharing_op, xen_domctl_mem_sharing_op, mem_sharing_op)	// depends

#if defined(XEN_DOMCTL_disable_migrate)
	XEN_DOMCTL(disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// input
#elif defined(XEN_DOMCTL_set_disable_migrate)
	XEN_DOMCTL(set_disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// input
#endif
#ifdef XEN_DOMCTL_query_disable_migrate
	XEN_DOMCTL_OUT(query_disable_migrate, xen_domctl_disable_migrate, disable_migrate)	// output
#endif
	XEN_DOMCTL_OUT(gettscinfo, xen_domctl_tsc_info, tsc_info)	// output
	XEN_DOMCTL(settscinfo, xen_domctl_tsc_info, tsc_info)	// input
	// TODO handle input and output of same field
	XEN_DOMCTL(getpageframeinfo3, xen_domctl_getpageframeinfo3, getpageframeinfo3)	// num, array input, array output
	XEN_DOMCTL(setvcpuextstate, xen_domctl_vcpuextstate, vcpuextstate)	// input
	XEN_DOMCTL_OUT(getvcpuextstate, xen_domctl_vcpuextstate, vcpuextstate)	// output, vcpu, size, xfeature_mask input
	XEN_DOMCTL(set_access_required, xen_domctl_set_access_required, access_required)	// input
	XEN_DOMCTL_OUT(get_runstate_info, xen_domctl_runstate_info, domain_runstate)	// output
	XEN_DOMCTL(setcorespersocket, xen_domctl_corespersocket, corespersocket)	// input
	XEN_DOMCTL_OUT(gdbsx_guestmemio, xen_domctl_gdbsx_memio, gdbsx_guest_memio)	// see structure, mostly input
	XEN_DOMCTL(gdbsx_pausevcpu, xen_domctl_gdbsx_pauseunp_vcpu, gdbsx_pauseunp_vcpu)	// input
	XEN_DOMCTL(gdbsx_unpausevcpu, xen_domctl_gdbsx_pauseunp_vcpu, gdbsx_pauseunp_vcpu)	// input
	XEN_DOMCTL_OUT(gdbsx_domstatus, xen_domctl_gdbsx_domstatus, gdbsx_domstatus)	// output
#undef XEN_DOMCTL_NOARG
#undef XEN_DOMCTL
#undef XEN_DOMCTL_OUT
	}

	if (type)
		type->print(status, &data->u, type);

	status_printf(status, "}");
}

static void
print_xen_domctl(print_status *status, const void *_data, const void *dummy)
{
	const struct xen_domctl *data = (const struct xen_domctl *)_data;
	if (status->entering)
		print_xen_domctl_enter(status, data);
	else
		print_xen_domctl_exit(status, data);
}

static const type_base type_xen_domctl = {
	{ print_xen_domctl, sizeof(STRUCT) }
};
#undef STRUCT

#ifdef IOCTL_PRIVCMD_MEMOP

#define STRUCT struct privcmd_memop
static void
print_privcmd_memop_enter(print_status *status, const struct privcmd_memop *data)
{
	const base_type* type = NULL;
	const char *name = NULL;

#define PRIVCMD_MEMOP_NOARG(_name) case XENMEM_ ## _name: \
	name = #_name; \
	break;
#define PRIVCMD_MEMOP(_name, struct, item) case XENMEM_ ## _name: \
	name = #_name; \
	type = &type_ ## struct.base; \
	break;
#define PRIVCMD_MEMOP_OUT(name, struct, item) PRIVCMD_MEMOP_NOARG(name)

	switch (data->cmd) {
	PRIVCMD_MEMOP(increase_reservation, xen_memory_reservation, reservation) // extent_start out
	PRIVCMD_MEMOP(decrease_reservation, xen_memory_reservation, reservation) // extent_start in
	PRIVCMD_MEMOP_NOARG(maximum_ram_page)
	PRIVCMD_MEMOP(current_reservation, domid, domid)
	PRIVCMD_MEMOP(maximum_reservation, domid, domid)
	PRIVCMD_MEMOP_OUT(machphys_mfn_list, xen_machphys_mfn_list, machphys_mfn_list)	// mostly out
	PRIVCMD_MEMOP(populate_physmap, xen_memory_reservation, reservation)	// extent_start in/out
	PRIVCMD_MEMOP(add_to_physmap, xen_add_to_physmap, add_to_physmap)	// input ??
	PRIVCMD_MEMOP_OUT(memory_map, xen_memory_map, memory_map)	// some input!
	PRIVCMD_MEMOP_OUT(machine_memory_map, xen_memory_map, memory_map)	// like memory_map
	PRIVCMD_MEMOP(exchange, xen_memory_exchange, exchange)	// some in some out
	PRIVCMD_MEMOP_OUT(machphys_mapping, xen_machphys_mapping, machphys_mapping)
	PRIVCMD_MEMOP(set_memory_map, xen_foreign_memory_map, foreign_memory_map)
	PRIVCMD_MEMOP_NOARG(maximum_gpfn)
	// defined in memory but no union for it in privcmd
//	PRIVCMD_MEMOP(set_pod_target, xen_pod_target, )	// input
//	PRIVCMD_MEMOP_OUT(get_pod_target, xen_pod_target, )
	PRIVCMD_MEMOP_NOARG(get_sharing_freed_pages)
#undef PRIVCMD_MEMOP_NOARG
#undef PRIVCMD_MEMOP
#undef PRIVCMD_MEMOP_OUT
	}

	if (name)
		status_printf(status, "{cmd=%s", name);
	else
		status_printf(status, "{cmd=UNKNOWN(%lu)", data->cmd);

	if (type)
		type->print(status, &data->u, type);
}

static void
print_privcmd_memop_exit(print_status *status, const struct privcmd_memop *data)
{
	const base_type* type = NULL;

#define PRIVCMD_MEMOP_NOARG(_name) case XENMEM_ ## _name: \
	break;
// print just end
#define PRIVCMD_MEMOP(name, struct, item) PRIVCMD_MEMOP_NOARG(name)
// print
#define PRIVCMD_MEMOP_OUT(_name, struct, item) case XENMEM_ ## _name: \
	type = &type_ ## struct.base; \
	break;

	switch (data->cmd) {
	PRIVCMD_MEMOP(increase_reservation, xen_memory_reservation, reservation) // extent_start out
	PRIVCMD_MEMOP(decrease_reservation, xen_memory_reservation, reservation) // extent_start in
	PRIVCMD_MEMOP_NOARG(maximum_ram_page)
	PRIVCMD_MEMOP(current_reservation, domid, domid)
	PRIVCMD_MEMOP(maximum_reservation, domid, domid)
	PRIVCMD_MEMOP_OUT(machphys_mfn_list, xen_machphys_mfn_list, machphys_mfn_list)	// mostly out
	PRIVCMD_MEMOP(populate_physmap, xen_memory_reservation, reservation)	// extent_start in/out
	PRIVCMD_MEMOP(add_to_physmap, xen_add_to_physmap, add_to_physmap)	// input ??
	PRIVCMD_MEMOP_OUT(memory_map, xen_memory_map, memory_map)	// some input!
	PRIVCMD_MEMOP_OUT(machine_memory_map, xen_memory_map, memory_map)	// like memory_map
	PRIVCMD_MEMOP(exchange, xen_memory_exchange, exchange)	// some in some out
	PRIVCMD_MEMOP_OUT(machphys_mapping, xen_machphys_mapping, machphys_mapping)
	PRIVCMD_MEMOP(set_memory_map, xen_foreign_memory_map, foreign_memory_map)
	PRIVCMD_MEMOP_NOARG(maximum_gpfn)
	// defined in memory but no union for it in privcmd
//	PRIVCMD_MEMOP(set_pod_target, xen_pod_target, )	// input
//	PRIVCMD_MEMOP_OUT(get_pod_target, xen_pod_target, )
	PRIVCMD_MEMOP_NOARG(get_sharing_freed_pages)
#undef PRIVCMD_MEMOP_NOARG
#undef PRIVCMD_MEMOP
#undef PRIVCMD_MEMOP_OUT
	}

	if (type)
		type->print(status, &data->u, type);

	status_printf(status, "}");
}

static void
print_privcmd_memop(print_status *status, const void *_data, const void *dummy)
{
	const struct privcmd_memop *data = (const struct privcmd_memop *)_data;
	if (status->entering)
		print_privcmd_memop_enter(status, data);
	else
		print_privcmd_memop_exit(status, data);
}

static const type_base type_privcmd_memop = {
	{ print_privcmd_memop, sizeof(STRUCT) }
};
#undef STRUCT

#endif

#include "hypercall.h"


// for arrays ??

static int
decode_struct0(const char *code_name, const base_type *type_input, const base_type *type_output, struct tcb *tcp, long arg)
{
	/* this assume all structures are maximum 256 bytes */
	char buf[256];
	print_status pr;
	const base_type *type;

	status_init(&pr, tcp);
	pr.entering = !!entering(tcp);

	if (entering(tcp)) {
		tprintf("%s, ", code_name);
		if (!verbose(tcp))
			tprintf("%#lx", arg);
	}
	if (!verbose(tcp))
		return 1;

	type = entering(tcp) ? type_input : type_output;
	if (!type)
		return 1;

	if (umoven(tcp, arg, type->size, buf) < 0) {
		tprintf("%#lx", arg);
		return 1;
	}
	if (!entering(tcp)) tprints("->");
	type->print(&pr, buf, type);
	status_flush(&pr);
	return 1;
}

#define decode_struct_input(code, struct) \
	case code: return decode_struct0(#code, &type_ ## struct.base, NULL, tcp, arg)
#define decode_struct_output(code, struct) \
	case code: return decode_struct0(#code, NULL, &type_ ## struct.base, tcp, arg)
#define decode_struct(code, istruct, ostruct) \
	case code: return decode_struct0(#code, &type_ ## istruct.base, &type_ ## ostruct.base, tcp, arg)

/*
 * Try to detect and handle the exact ioctl code
 * return 1 if handled, 0 otherwise
 */
int ioctl_decode_exact(struct tcb *tcp)
{
	long code = tcp->u_arg[1];
	long arg =  tcp->u_arg[2];

	switch (code) {
	/* Xen event channels */
#ifdef IOCTL_EVTCHN_BIND_VIRQ
	decode_struct_input(IOCTL_EVTCHN_BIND_VIRQ, evtchn_bind_virq);
#endif
#ifdef IOCTL_EVTCHN_BIND_INTERDOMAIN
	decode_struct_input(IOCTL_EVTCHN_BIND_INTERDOMAIN, evtchn_bind_interdomain);
#endif
#ifdef IOCTL_EVTCHN_BIND_UNBOUND_PORT
	decode_struct_input(IOCTL_EVTCHN_BIND_UNBOUND_PORT, evtchn_bind_unbound_port);
#endif
#ifdef IOCTL_EVTCHN_UNBIND
	decode_struct_input(IOCTL_EVTCHN_UNBIND, evtchn_unbind);
#endif
#ifdef IOCTL_EVTCHN_NOTIFY
	decode_struct_input(IOCTL_EVTCHN_NOTIFY, evtchn_notify);
#endif
#ifdef IOCTL_EVTCHN_RESET
	case IOCTL_EVTCHN_RESET:
		if (entering(tcp))
			tprintf("IOCTL_EVTCHN_RESET, %#lx", arg);
		return 1;
#endif

	/* Xen privcmd */
#ifdef IOCTL_PRIVCMD_HYPERCALL
	decode_struct_input(IOCTL_PRIVCMD_HYPERCALL, hypercall);
#endif
#ifdef IOCTL_PRIVCMD_MMAP
	decode_struct_input(IOCTL_PRIVCMD_MMAP, privcmd_mmap);
#endif
#ifdef IOCTL_PRIVCMD_MMAPBATCH
	decode_struct_output(IOCTL_PRIVCMD_MMAPBATCH, privcmd_mmapbatch);
#endif
#ifdef IOCTL_PRIVCMD_MMAPBATCH_V2
	decode_struct_output(IOCTL_PRIVCMD_MMAPBATCH_V2, privcmd_mmapbatch_v2);
#endif

	/* XenServer Specific */
#ifdef IOCTL_EVTCHN_RESTRICT_DOMID
	decode_struct_input(IOCTL_EVTCHN_RESTRICT_DOMID, evtchn_restrict_domid);
#endif
#ifdef IOCTL_PRIVCMD_HVMOP
	decode_struct_output(IOCTL_PRIVCMD_HVMOP, privcmd_hvmop);
#endif
#ifdef IOCTL_PRIVCMD_RESTRICT_DOMID
	decode_struct_input(IOCTL_PRIVCMD_RESTRICT_DOMID, privcmd_restrict_domid);
#endif
#ifdef IOCTL_PRIVCMD_SHUTDOWN
	decode_struct_input(IOCTL_PRIVCMD_SHUTDOWN, sched_remote_shutdown);
#endif
#ifdef IOCTL_PRIVCMD_MEMOP
	decode_struct(IOCTL_PRIVCMD_MEMOP, privcmd_memop, privcmd_memop);
#endif
#ifdef IOCTL_PRIVCMD_DOMCTL
	decode_struct(IOCTL_PRIVCMD_DOMCTL, xen_domctl, xen_domctl);
#endif
	}
	return 0;
}

