/*
 * hold information on status
 * need flush if space is not enough
 */
typedef struct {
	char *dest, *dest_end;
	char buffer[256];
	struct tcb *tcp;
	char entering;	/* != 0 if entering */
} print_status;

static inline void
status_init(print_status *status, struct tcb *tcp)
{
	status->dest     = status->buffer;
	status->dest_end = status->buffer + sizeof(status->buffer);
	status->tcp      = tcp;
}

void status_printf(print_status *status, const char *fmt, ...);
void status_flush(print_status *status);

/* Define structure to handle print of a structure */
typedef void (*print_func_t)(print_status *status, const void *data, const void *info);

#define DATA_START base_type base

/*
 * Define base types 
 */
typedef struct {
	print_func_t print;
	unsigned size;
} base_type;

typedef struct {
	DATA_START;
} type_base;

const type_base type_void;
const type_base type_guid;

#define DEFINE_BASE_TYPE(name, type, fmt) \
static void print_ ## name(print_status *status, const void *data, const void *info) { \
	status_printf(status, fmt, *((type*)(data))); \
} \
static const type_base type_ ## name = { { print_ ## name, sizeof(type) } };

/*
 * Define structures
 */
typedef struct {
	const char *name;	/* name of field NULL if last */
	const base_type* info;	/* information for printing */
	size_t off;		/* offset into structure */
	// size ??
} struct_item;

typedef struct {
	DATA_START;
	const struct_item *items;
} type_struct;

void print_struct(print_status *status, const void *_data, const void *_info);

#define START_STRUCT(name) \
static const struct_item name ## _items[] = {

#define BASE_ITEM(type, name) \
	{ #name, &type_ ## type.base, offsetof(STRUCT, name) },

#define END_STRUCT(name) \
	{ NULL, } \
}; \
	static const type_struct type_ ## name = { \
		{ print_struct, sizeof(STRUCT) }, \
		name ## _items \
	};


/*
 * Fixed arrays
 * Define structure and macros for arrays with fixed length
 */
typedef struct {
	DATA_START;
	const base_type *item_type;
} type_fixed_array;

void print_fixed_array(print_status *status, const void *_data, const void *_info);

#define FIXED_ARRAY(name, field, base_type) \
	static const type_fixed_array type_ ## name = { \
		{ print_fixed_array, sizeof(((STRUCT*)0)->field) }, \
		&type_ ## base_type.base \
	};


/*
 * Define arrays with size specified
 */
typedef struct {
	DATA_START;
	unsigned num_size;
	const base_type *item_type;
	ssize_t num_off;	/* offset to number from data */
} type_ptr_array;

void print_ptr_array(print_status *status, const void *_data, const void *_info);

#define PTR_ARRAY(name, field, base_type, count_field) \
	static const type_ptr_array type_## name = { \
		{ print_ptr_array, sizeof(void*) }, \
		sizeof(((STRUCT*)0)->count_field), \
		&type_ ## base_type.base, \
		offsetof(STRUCT, count_field) - offsetof(STRUCT, field) \
	};


/*
 * Define arrays of unsigned longs
 * This support detecting repetitions and scales inside the sequence
 */
typedef struct {
	DATA_START;
	unsigned num_size;
	ssize_t num_off;	/* offset to number from data */
} type_ulong_array;

void print_ulong_array(print_status *status, const void *_data, const void *_info);

