/*
 * Base types
 */
DEFINE_BASE_TYPE(int, int, "%d");
DEFINE_BASE_TYPE(uint, unsigned int, "%u");
DEFINE_BASE_TYPE(ulong, unsigned long, "%lu");
DEFINE_BASE_TYPE(domid, domid_t, "%u");
DEFINE_BASE_TYPE(u8,  uint8_t,  "%u");
DEFINE_BASE_TYPE(u16, uint16_t, "%u");
DEFINE_BASE_TYPE(u32, uint32_t, "%u");
DEFINE_BASE_TYPE(u64, __u64, "%" PRIu64);
DEFINE_BASE_TYPE(p64, __u64, "%#" PRIx64);	// TODO for now just print
/* typedef long unsigned int, xen_pfn_t; */
DEFINE_BASE_TYPE(xen_pfn_t, unsigned long, "%lu");


/*
 * Event channels
 */
#define STRUCT struct ioctl_evtchn_bind_virq
START_STRUCT(evtchn_bind_virq)
	BASE_ITEM(uint, virq)
END_STRUCT(evtchn_bind_virq)
#undef STRUCT

#define STRUCT struct ioctl_evtchn_bind_interdomain
START_STRUCT(evtchn_bind_interdomain)
	BASE_ITEM(uint, remote_domain)
	BASE_ITEM(uint, remote_port)
END_STRUCT(evtchn_bind_interdomain)
#undef STRUCT

#define STRUCT struct ioctl_evtchn_bind_unbound_port
START_STRUCT(evtchn_bind_unbound_port)
	BASE_ITEM(uint, remote_domain)
END_STRUCT(evtchn_bind_unbound_port)
#undef STRUCT

#define STRUCT struct ioctl_evtchn_unbind
START_STRUCT(evtchn_unbind)
	BASE_ITEM(uint, port)
END_STRUCT(evtchn_unbind)
#undef STRUCT

#define STRUCT struct ioctl_evtchn_notify
START_STRUCT(evtchn_notify)
	BASE_ITEM(uint, port)
END_STRUCT(evtchn_notify)
#undef STRUCT

#ifdef IOCTL_EVTCHN_RESTRICT_DOMID
#define STRUCT struct ioctl_evtchn_restrict_domid
START_STRUCT(evtchn_restrict_domid)
	BASE_ITEM(domid, domid)
END_STRUCT(evtchn_restrict_domid)
#undef STRUCT
#endif


/*
 * privcmd
 */
#define STRUCT privcmd_hypercall_t
FIXED_ARRAY(privcmd_hypercall_args, arg, u64)

START_STRUCT(privcmd_hypercall)
	BASE_ITEM(u64, op)
	BASE_ITEM(privcmd_hypercall_args, arg)
END_STRUCT(privcmd_hypercall)
#undef STRUCT

#if 0
#define STRUCT struct xen_hvm_set_pci_intx_level
START_STRUCT(xen_hvm_set_pci_intx_level)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u8, domain)
	BASE_ITEM(u8, bus)
	BASE_ITEM(u8, device)
	BASE_ITEM(u8, intx)
	BASE_ITEM(u8, level)
END_STRUCT(xen_hvm_set_pci_intx_level)
#undef STRUCT

#define STRUCT struct xen_hvm_set_isa_irq_level
START_STRUCT(xen_hvm_set_isa_irq_level)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u8, isa_irq)
	BASE_ITEM(u8, level)
END_STRUCT(xen_hvm_set_isa_irq_level)
#undef STRUCT

#define STRUCT struct xen_hvm_set_pci_link_route
START_STRUCT(xen_hvm_set_pci_link_route)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u8, link)
	BASE_ITEM(u8, isa_irq)
END_STRUCT(xen_hvm_set_pci_link_route)
#undef STRUCT
#endif

#define STRUCT struct xen_hvm_param
START_STRUCT(xen_hvm_param)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u32, index)
	BASE_ITEM(u64, value)
END_STRUCT(xen_hvm_param)
#undef STRUCT

#if 0
#define STRUCT struct xen_hvm_modified_memory
START_STRUCT(xen_hvm_modified_memory)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u64, first_pfn)
	BASE_ITEM(u64, nr)
END_STRUCT(xen_hvm_modified_memory)
#undef STRUCT

#define STRUCT struct xen_hvm_set_mem_type
START_STRUCT(xen_hvm_set_mem_type)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u16, hvmmem_type)
	BASE_ITEM(u32, nr)
	BASE_ITEM(u64, first_pfn)
END_STRUCT(xen_hvm_set_mem_type)
#undef STRUCT
#endif

#define STRUCT struct xen_hvm_pagetable_dying
FIXED_ARRAY(xen_hvm_pagetable_dying_pads, pad, u16)

START_STRUCT(xen_hvm_pagetable_dying)
	BASE_ITEM(domid, domid)
	BASE_ITEM(xen_hvm_pagetable_dying_pads, pad)
	BASE_ITEM(u64, gpa)
END_STRUCT(xen_hvm_pagetable_dying)
#undef STRUCT

#define STRUCT struct xen_hvm_get_time
START_STRUCT(xen_hvm_get_time)
	BASE_ITEM(u64, now)
END_STRUCT(xen_hvm_get_time)
#undef STRUCT

#if XEN_SYSCTL_INTERFACE_VERSION < 0x0000000B
/*
 * Gross hack, to counter a gross API breakage.  Xen 4.5 deprecated the
 * HVMOP_{get,set}_mem_access hypercalls, removing their structure definitions
 * but not hypercall numbers.
 *
 * While the sysctl interface version is unrelated to hvmops, it serves as a
 * Xen-4.4 vs Xen-4.5 gate.
 */
#define STRUCT struct xen_hvm_set_mem_access
START_STRUCT(xen_hvm_set_mem_access)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u16, hvmmem_access)
	BASE_ITEM(u32, nr)
	BASE_ITEM(u64, first_pfn)
END_STRUCT(xen_hvm_set_mem_access)
#undef STRUCT

#define STRUCT struct xen_hvm_get_mem_access
START_STRUCT(xen_hvm_get_mem_access)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u16, hvmmem_access)
	BASE_ITEM(u64, pfn)
END_STRUCT(xen_hvm_get_mem_access)
#undef STRUCT
#endif /* XEN_SYSCTL_INTERFACE_VERSION < 0x0000000B */

#if 0
#define STRUCT struct xen_hvm_inject_trap
START_STRUCT(xen_hvm_inject_trap)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u32, vcpuid)
	BASE_ITEM(u32, error_code)
	BASE_ITEM(u64, cr2)
END_STRUCT(xen_hvm_inject_trap)
#undef STRUCT
#endif

#define STRUCT struct xen_hvm_get_mem_type
FIXED_ARRAY(xen_hvm_get_mem_type_pads, pad, u16)

START_STRUCT(xen_hvm_get_mem_type)
	BASE_ITEM(domid, domid)
	BASE_ITEM(u16, mem_type)
	BASE_ITEM(xen_hvm_get_mem_type_pads, pad)
	BASE_ITEM(u64, pfn)
END_STRUCT(xen_hvm_get_mem_type)
#undef STRUCT

#ifdef HVMOP_set_driver_version
#define STRUCT struct xen_hvm_set_driver_version
START_STRUCT(xen_hvm_set_driver_version)
	BASE_ITEM(u32, build)
END_STRUCT(xen_hvm_set_driver_version)
#undef STRUCT
#endif /* HVMOP_set_driver_version */

#define STRUCT struct xen_hvm_xentrace
FIXED_ARRAY(xen_hvm_xentrace_extras, extra, u8)

START_STRUCT(xen_hvm_xentrace)
	BASE_ITEM(u16, event)
	BASE_ITEM(u16, extra_bytes)
	BASE_ITEM(xen_hvm_xentrace_extras, extra)
END_STRUCT(xen_hvm_xentrace)
#undef STRUCT

#ifdef IOCTL_PRIVCMD_SHUTDOWN
#define STRUCT struct sched_remote_shutdown
START_STRUCT(sched_remote_shutdown)
	BASE_ITEM(domid, domain_id)
	BASE_ITEM(uint, reason)
END_STRUCT(sched_remote_shutdown)
#undef STRUCT
#endif

#ifdef IOCTL_PRIVCMD_RESTRICT_DOMID
#define STRUCT struct privcmd_restrict_domid
START_STRUCT(privcmd_restrict_domid)
	BASE_ITEM(domid, domid)
END_STRUCT(privcmd_restrict_domid)
#undef STRUCT
#endif

#define STRUCT struct privcmd_mmap_entry
START_STRUCT(privcmd_mmap_entry)
	BASE_ITEM(u64, va)
	BASE_ITEM(u64, mfn)
	BASE_ITEM(u64, npages)
END_STRUCT(privcmd_mmap_entry)
#undef STRUCT

#define STRUCT struct privcmd_mmap
PTR_ARRAY(privcmd_mmap_entries, entry, privcmd_mmap_entry, num)

START_STRUCT(privcmd_mmap)
	BASE_ITEM(int, num)
	BASE_ITEM(domid, dom)
	BASE_ITEM(privcmd_mmap_entries, entry)
END_STRUCT(privcmd_mmap)
#undef STRUCT

#define STRUCT struct privcmd_mmapbatch
static const type_ulong_array type_privcmd_mmapbatch_arr = {
	{ print_ulong_array, sizeof(void*) },
	sizeof(((struct privcmd_mmapbatch*)0)->num),
	offsetof(struct privcmd_mmapbatch, num) - offsetof(struct privcmd_mmapbatch, arr)
};
START_STRUCT(privcmd_mmapbatch)
	BASE_ITEM(int, num)
	BASE_ITEM(domid, dom)
	BASE_ITEM(p64, addr)
	BASE_ITEM(privcmd_mmapbatch_arr, arr)
END_STRUCT(privcmd_mmapbatch)
#undef STRUCT

#define STRUCT struct privcmd_mmapbatch_v2
static const type_ulong_array type_privcmd_mmapbatch_v2_arr = {
	{ print_ulong_array, sizeof(void*) },
	sizeof(((struct privcmd_mmapbatch_v2*)0)->num),
	offsetof(struct privcmd_mmapbatch_v2, num) - offsetof(struct privcmd_mmapbatch_v2, arr)
};
PTR_ARRAY(privcmd_mmapbatch_v2_err, err, int, num)

START_STRUCT(privcmd_mmapbatch_v2)
	BASE_ITEM(uint, num)
	BASE_ITEM(domid, dom)
	BASE_ITEM(p64, addr)
	BASE_ITEM(privcmd_mmapbatch_v2_arr, arr)
	BASE_ITEM(privcmd_mmapbatch_v2_err, err)
END_STRUCT(privcmd_mmapbatch_v2)
#undef STRUCT

#define STRUCT struct xen_domctl_createdomain
START_STRUCT(xen_domctl_createdomain)
	BASE_ITEM(u32, ssidref)
	BASE_ITEM(guid, handle)
	BASE_ITEM(u32, flags)
END_STRUCT(xen_domctl_createdomain)
#undef STRUCT

#define STRUCT struct xen_domctl_getdomaininfo
START_STRUCT(xen_domctl_getdomaininfo)
	BASE_ITEM(domid, domain)
	BASE_ITEM(u32, flags)
	BASE_ITEM(u64, tot_pages)
	BASE_ITEM(u64, max_pages)
	BASE_ITEM(u64, outstanding_pages)
	BASE_ITEM(u64, shr_pages)
	BASE_ITEM(u64, paged_pages)
	BASE_ITEM(u64, shared_info_frame)
	BASE_ITEM(u64, cpu_time)
	BASE_ITEM(u32, nr_online_vcpus)
	BASE_ITEM(u32, max_vcpu_id)
	BASE_ITEM(u32, ssidref)
	BASE_ITEM(guid, handle)
	BASE_ITEM(u32, cpupool)
END_STRUCT(xen_domctl_getdomaininfo)
#undef STRUCT

#define STRUCT struct xenctl_bitmap
START_STRUCT(xenctl_bitmap)
	BASE_ITEM(p64, bitmap)
	BASE_ITEM(u32, nr_bits)
END_STRUCT(xenctl_bitmap)
#undef STRUCT

#define STRUCT struct xen_domctl_vcpuaffinity
START_STRUCT(xen_domctl_vcpuaffinity)
	BASE_ITEM(u32, vcpu)
#ifndef XEN_VCPUAFFINITY_HARD
/* Xen 4.4 and older.  Just a single bitmap */
	BASE_ITEM(xenctl_bitmap, cpumap)
#else
/* Xen 4.5 - Added a soft vs hard split */
	BASE_ITEM(u32, flags)
	BASE_ITEM(xenctl_bitmap, cpumap_hard)
	BASE_ITEM(xenctl_bitmap, cpumap_soft)
#endif
END_STRUCT(xen_domctl_vcpuaffinity)
#undef STRUCT

#define STRUCT struct xen_domctl_max_mem
START_STRUCT(xen_domctl_max_mem)
	BASE_ITEM(u64, max_memkb)
END_STRUCT(xen_domctl_max_mem)
#undef STRUCT

#define STRUCT struct xen_domctl_vcpucontext
START_STRUCT(xen_domctl_vcpucontext)
	BASE_ITEM(u32, vcpu)
	BASE_ITEM(p64, ctxt)
END_STRUCT(xen_domctl_vcpucontext)
#undef STRUCT

#ifdef XEN_DOMCTL_max_vcpus
#define STRUCT struct xen_domctl_max_vcpus
START_STRUCT(xen_domctl_max_vcpus)
	BASE_ITEM(u32, max)
END_STRUCT(xen_domctl_max_vcpus)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_setdomainhandle
START_STRUCT(xen_domctl_setdomainhandle)
	BASE_ITEM(guid, handle)
END_STRUCT(xen_domctl_setdomainhandle)
#undef STRUCT

#define STRUCT struct xen_domctl_setdebugging
START_STRUCT(xen_domctl_setdebugging)
	BASE_ITEM(u8, enable)
END_STRUCT(xen_domctl_setdebugging)
#undef STRUCT

#define STRUCT struct xen_domctl_irq_permission
START_STRUCT(xen_domctl_irq_permission)
	BASE_ITEM(u8, pirq)
	BASE_ITEM(u8, allow_access)
END_STRUCT(xen_domctl_irq_permission)
#undef STRUCT

#define STRUCT struct xen_domctl_iomem_permission
START_STRUCT(xen_domctl_iomem_permission)
	BASE_ITEM(u64, first_mfn)
	BASE_ITEM(u64, nr_mfns)
	BASE_ITEM(u8, allow_access)
END_STRUCT(xen_domctl_iomem_permission)
#undef STRUCT

#define STRUCT struct xen_domctl_ioport_permission
START_STRUCT(xen_domctl_ioport_permission)
	BASE_ITEM(u32, first_port)
	BASE_ITEM(u32, nr_ports)
	BASE_ITEM(u8, allow_access)
END_STRUCT(xen_domctl_ioport_permission)
#undef STRUCT

#define STRUCT struct xen_domctl_hypercall_init
START_STRUCT(xen_domctl_hypercall_init)
	BASE_ITEM(u64, gmfn)
END_STRUCT(xen_domctl_hypercall_init)
#undef STRUCT

#define STRUCT struct xen_domctl_settimeoffset
START_STRUCT(xen_domctl_settimeoffset)
	BASE_ITEM(int, time_offset_seconds)
END_STRUCT(xen_domctl_settimeoffset)
#undef STRUCT

#define STRUCT struct xen_domctl_sendtrigger
START_STRUCT(xen_domctl_sendtrigger)
	BASE_ITEM(u32, trigger)
	BASE_ITEM(u32, vcpu)
END_STRUCT(xen_domctl_sendtrigger)
#undef STRUCT

#define STRUCT struct xen_domctl_subscribe
START_STRUCT(xen_domctl_subscribe)
	BASE_ITEM(u32, port)
END_STRUCT(xen_domctl_subscribe)
#undef STRUCT

#define STRUCT struct xen_domctl_hvmcontext
START_STRUCT(xen_domctl_hvmcontext)
	BASE_ITEM(u32, size)
	BASE_ITEM(p64, buffer)
END_STRUCT(xen_domctl_hvmcontext)
#undef STRUCT

#define STRUCT struct xen_domctl_address_size
START_STRUCT(xen_domctl_address_size)
	BASE_ITEM(u32, size)
END_STRUCT(xen_domctl_address_size)
#undef STRUCT

#define STRUCT struct xen_domctl_assign_device
START_STRUCT(xen_domctl_assign_device)
#ifndef XEN_DOMCTL_DEV_DT
/* Xen 4.5 and older.  Just a single sbdf */
	BASE_ITEM(u32, machine_sbdf)
#else
/* Xen 4.6 and newer.  Split for ARM */
        BASE_ITEM(u32, dev)
/* TODO: how to represent the union? */
#endif
END_STRUCT(xen_domctl_assign_device)
#undef STRUCT

#define STRUCT struct xen_domctl_memory_mapping
START_STRUCT(xen_domctl_memory_mapping)
	BASE_ITEM(u64, first_gfn)
	BASE_ITEM(u64, first_mfn)
	BASE_ITEM(u64, nr_mfns)
	BASE_ITEM(u32, add_mapping)
	BASE_ITEM(u32, padding)
END_STRUCT(xen_domctl_memory_mapping)
#undef STRUCT

#define STRUCT struct xen_domctl_ioport_mapping
START_STRUCT(xen_domctl_ioport_mapping)
	BASE_ITEM(u32, first_gport)
	BASE_ITEM(u32, first_mport)
	BASE_ITEM(u32, nr_ports)
	BASE_ITEM(u32, add_mapping)
END_STRUCT(xen_domctl_ioport_mapping)
#undef STRUCT

#ifdef XEN_DOMCTL_pin_mem_cacheattr
#define STRUCT struct xen_domctl_pin_mem_cacheattr
START_STRUCT(xen_domctl_pin_mem_cacheattr)
	BASE_ITEM(u64, start)
	BASE_ITEM(u64, end)
	BASE_ITEM(u32, type)
END_STRUCT(xen_domctl_pin_mem_cacheattr)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_ext_vcpucontext
START_STRUCT(xen_domctl_ext_vcpucontext)
	BASE_ITEM(u32, vcpu)
	BASE_ITEM(u32, size)
	BASE_ITEM(u64, syscall32_callback_eip)
	BASE_ITEM(u64, sysenter_callback_eip)
	BASE_ITEM(u16, syscall32_callback_cs)
	BASE_ITEM(u16, sysenter_callback_cs)
	BASE_ITEM(u8, syscall32_disables_events)
	BASE_ITEM(u8, sysenter_disables_events)
END_STRUCT(xen_domctl_ext_vcpucontext)
#undef STRUCT

#define STRUCT struct xen_domctl_set_target
START_STRUCT(xen_domctl_set_target)
	BASE_ITEM(domid, target)
END_STRUCT(xen_domctl_set_target)
#undef STRUCT

#ifdef XEN_DOMCTL_set_cpuid
#define STRUCT struct xen_domctl_cpuid
FIXED_ARRAY(xen_domctl_cpuid_inputs, input, u32)

START_STRUCT(xen_domctl_cpuid)
	BASE_ITEM(xen_domctl_cpuid_inputs, input)
	BASE_ITEM(u32, eax)
	BASE_ITEM(u32, ebx)
	BASE_ITEM(u32, ecx)
	BASE_ITEM(u32, edx)
END_STRUCT(xen_domctl_cpuid)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_debug_op
START_STRUCT(xen_domctl_debug_op)
	BASE_ITEM(u32, op)
	BASE_ITEM(u32, vcpu)
END_STRUCT(xen_domctl_debug_op)
#undef STRUCT

#ifdef XEN_DOMCTL_MEM_EVENT_OP_SHARING
#define STRUCT struct xen_domctl_mem_event_op
START_STRUCT(xen_domctl_mem_event_op)
	BASE_ITEM(u32, op)
	BASE_ITEM(u32, mode)
	BASE_ITEM(u32, port)
END_STRUCT(xen_domctl_mem_event_op)
#undef STRUCT
#endif

#ifdef XEN_DOMCTL_VM_EVENT_OP_SHARING
#define STRUCT struct xen_domctl_vm_event_op
START_STRUCT(xen_domctl_vm_event_op)
	BASE_ITEM(u32, op)
	BASE_ITEM(u32, mode)
#ifndef XEN_VM_EVENT_GET_VERSION /* Xen <= 4.12 */
	BASE_ITEM(u32, port)
#else
	BASE_ITEM(u32, u.enable.port)
#endif
END_STRUCT(xen_domctl_vm_event_op)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_disable_migrate
START_STRUCT(xen_domctl_disable_migrate)
	BASE_ITEM(u32, disable)
END_STRUCT(xen_domctl_disable_migrate)
#undef STRUCT

#if XEN_DOMCTL_INTERFACE_VERSION < 0x0000000B
#define STRUCT struct xen_guest_tsc_info
START_STRUCT(xen_guest_tsc_info)
	BASE_ITEM(u32, tsc_mode)
	BASE_ITEM(u32, gtsc_khz)
	BASE_ITEM(u32, incarnation)
	BASE_ITEM(u32, pad)
	BASE_ITEM(u64, elapsed_nsec)
END_STRUCT(xen_guest_tsc_info)
#undef STRUCT

#define STRUCT struct xen_domctl_tsc_info
START_STRUCT(xen_domctl_tsc_info)
	BASE_ITEM(p64, out_info)
	BASE_ITEM(xen_guest_tsc_info, info)
END_STRUCT(xen_domctl_tsc_info)
#undef STRUCT

#else

#define STRUCT struct xen_domctl_tsc_info
START_STRUCT(xen_domctl_tsc_info)
	BASE_ITEM(u32, tsc_mode)
	BASE_ITEM(u32, gtsc_khz)
	BASE_ITEM(u32, incarnation)
	BASE_ITEM(u32, pad)
	BASE_ITEM(u64, elapsed_nsec)
END_STRUCT(xen_domctl_tsc_info)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_vcpuextstate
START_STRUCT(xen_domctl_vcpuextstate)
	BASE_ITEM(u32, vcpu)
	BASE_ITEM(u64, xfeature_mask)
	BASE_ITEM(u64, size)
	BASE_ITEM(p64, buffer)
END_STRUCT(xen_domctl_vcpuextstate)
#undef STRUCT

#define STRUCT struct xen_domctl_set_access_required
START_STRUCT(xen_domctl_set_access_required)
	BASE_ITEM(u8, access_required)
END_STRUCT(xen_domctl_set_access_required)
#undef STRUCT

#define STRUCT struct xen_domctl_corespersocket
START_STRUCT(xen_domctl_corespersocket)
	BASE_ITEM(u32, cores_per_socket)
END_STRUCT(xen_domctl_corespersocket)
#undef STRUCT

#define STRUCT struct xen_domctl_gdbsx_pauseunp_vcpu
START_STRUCT(xen_domctl_gdbsx_pauseunp_vcpu)
	BASE_ITEM(u32, vcpu)
END_STRUCT(xen_domctl_gdbsx_pauseunp_vcpu)
#undef STRUCT

#define STRUCT struct xen_domctl_gdbsx_domstatus
START_STRUCT(xen_domctl_gdbsx_domstatus)
	BASE_ITEM(u8, paused)
	BASE_ITEM(u32, vcpu_id)
	BASE_ITEM(u32, vcpu_ev)
END_STRUCT(xen_domctl_gdbsx_domstatus)
#undef STRUCT

#define STRUCT struct xen_domctl_runstate_info
FIXED_ARRAY(xen_domctl_runstate_info_times, time, u64)

START_STRUCT(xen_domctl_runstate_info)
	BASE_ITEM(u32, state)
	BASE_ITEM(u32, missed_changes)
	BASE_ITEM(u64, state_entry_time)
	BASE_ITEM(xen_domctl_runstate_info_times, time)
END_STRUCT(xen_domctl_runstate_info)
#undef STRUCT

#define STRUCT struct xen_domctl_gdbsx_memio
START_STRUCT(xen_domctl_gdbsx_memio)
	BASE_ITEM(u64, pgd3val)
	BASE_ITEM(u64, gva)
	BASE_ITEM(u64, uva)
	BASE_ITEM(u32, len)
	BASE_ITEM(u8, gwr)
	BASE_ITEM(u32, remain)
END_STRUCT(xen_domctl_gdbsx_memio)
#undef STRUCT

#ifdef XEN_DOMCTL_getpageframeinfo
#define STRUCT struct xen_domctl_getpageframeinfo
START_STRUCT(xen_domctl_getpageframeinfo)
	BASE_ITEM(u64, gmfn)
	BASE_ITEM(u32, type)
END_STRUCT(xen_domctl_getpageframeinfo)
#undef STRUCT
#endif

#ifdef XEN_DOMCTL_getpageframeinfo2
#define STRUCT struct xen_domctl_getpageframeinfo2
START_STRUCT(xen_domctl_getpageframeinfo2)
	BASE_ITEM(u64, num)
	BASE_ITEM(p64, array)
END_STRUCT(xen_domctl_getpageframeinfo2)
#undef STRUCT
#endif

#define STRUCT struct xen_domctl_getvcpuinfo
START_STRUCT(xen_domctl_getvcpuinfo)
	BASE_ITEM(u32, vcpu)
	BASE_ITEM(u8, online)
	BASE_ITEM(u8, blocked)
	BASE_ITEM(u8, running)
	BASE_ITEM(u64, cpu_time)
	BASE_ITEM(u32, cpu)
END_STRUCT(xen_domctl_getvcpuinfo)
#undef STRUCT

#define STRUCT struct xen_domctl_get_device_group
START_STRUCT(xen_domctl_get_device_group)
	BASE_ITEM(u32, machine_sbdf)
	BASE_ITEM(u32, max_sdevs)
	BASE_ITEM(u32, num_sdevs)
	BASE_ITEM(p64, sdev_array)
END_STRUCT(xen_domctl_get_device_group)
#undef STRUCT

#define STRUCT struct xen_domctl_hvmcontext_partial
START_STRUCT(xen_domctl_hvmcontext_partial)
	BASE_ITEM(u32, type)
	BASE_ITEM(u32, instance)
	BASE_ITEM(p64, buffer)
END_STRUCT(xen_domctl_hvmcontext_partial)
#undef STRUCT

#define STRUCT struct xen_domctl_getpageframeinfo3
START_STRUCT(xen_domctl_getpageframeinfo3)
	BASE_ITEM(u64, num)
	BASE_ITEM(p64, array)
END_STRUCT(xen_domctl_getpageframeinfo3)
#undef STRUCT


/*
 * Memory operations
 */
#define STRUCT struct xen_memory_reservation
START_STRUCT(xen_memory_reservation)
	BASE_ITEM(p64, extent_start)	// pointer to xen_pfn_t
	BASE_ITEM(ulong, nr_extents)
	BASE_ITEM(uint, extent_order)
	BASE_ITEM(uint, mem_flags)
	BASE_ITEM(domid, domid)
END_STRUCT(xen_memory_reservation)
#undef STRUCT

#define STRUCT struct xen_memory_exchange
START_STRUCT(xen_memory_exchange)
	BASE_ITEM(xen_memory_reservation, in)
	BASE_ITEM(xen_memory_reservation, out)
	BASE_ITEM(ulong, nr_exchanged)
END_STRUCT(xen_memory_exchange)
#undef STRUCT

#define STRUCT struct xen_add_to_physmap
START_STRUCT(xen_add_to_physmap)
	BASE_ITEM(domid, domid)
	BASE_ITEM(uint, space)
	BASE_ITEM(ulong, idx)
	BASE_ITEM(xen_pfn_t, gpfn)
END_STRUCT(xen_add_to_physmap)
#undef STRUCT

#define STRUCT struct xen_memory_map
START_STRUCT(xen_memory_map)
	BASE_ITEM(uint, nr_entries)
	BASE_ITEM(p64, buffer)	// pointer to void
END_STRUCT(xen_memory_map)
#undef STRUCT

#define STRUCT struct xen_foreign_memory_map
START_STRUCT(xen_foreign_memory_map)
	BASE_ITEM(domid, domid)
	BASE_ITEM(xen_memory_map, map)
END_STRUCT(xen_foreign_memory_map)
#undef STRUCT

#define STRUCT struct xen_machphys_mfn_list
START_STRUCT(xen_machphys_mfn_list)
	BASE_ITEM(uint, max_extents)
	BASE_ITEM(p64, extent_start)	// pointer to xen_pfn_t
	BASE_ITEM(uint, nr_extents)
END_STRUCT(xen_machphys_mfn_list)
#undef STRUCT

#define STRUCT struct xen_machphys_mapping
START_STRUCT(xen_machphys_mapping)
	BASE_ITEM(ulong, v_start)
	BASE_ITEM(ulong, v_end)
	BASE_ITEM(ulong, max_mfn)
END_STRUCT(xen_machphys_mapping)
#undef STRUCT

